// @ myth - client - screen.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// Screen I/O.
//
// This file contain three classes, TTY, TERMCAP and SCREEN.
// There may at most exists one object of the class TTY and one object
// of class TERMCAP or SCREEN at any time.
//
// TTY set up the screen driver when defined:
//    TTY tty(fd);               // fd must be connected to the tty.
//
// When the tty is deleted (or goes out of scope) the screen will be
// reset.
//
// The special keys may be exmained by:
//    tty.EraseChar(), tty.EraseLine(), tty.EraseWord(),
//    tty.RefreshChar(), tty.SuspendChar() and tty.QuitChar().
// Which all return the char used for this function.
//
// TERMCAP provide a low level interface to the termcap database.
// It is initialised by:
//    TERMCAP termcap;
//
// Now the functions in the termcap database can be used:
//    coor size = termcap.size(); // Return the size of the screen
//    cursor(coor(3, 6));         // Move cursor to (3, 6)
//    etc...
//
// To minimize the need for output to the screen, you can use the
// class SCREEN. It keep track an the current content of the screen,
// and the current position of the cursor, to avoid sending
// unnnecesary information to the screen. It also provide a way to
// temporarily cover part of the screen with a picture.
//    SCREEN screen;

// @@ include

#include "client/screen.h"
#include "common/error.h"
#include "common/format.h"
#include <ctype.h>

extern char PC;			// Padding character
extern char* BC;		// Backspace chareter
extern char* UP;		// Cursor up
extern short ospeed;		// Terminal output speed

// @@ const

// Determine when we will use TERMCAP::ClearScreen.
static const int MagicWindowCost = 5;
static const int MagicLineCost = 10;

// @@ static

overload wrong;

static void wrong(const char* what);
static void wrong(STRING what);
static void OutPut(char c);

// @@ * class TTY

// @@@ TTY::TTY(const char* name)

TTY::TTY(const char* name) : (BUFFER_RDWR, name)
{
    if(!GoodInput() || !GoodOutput())
	wrong(format("Can't open %s", name));

    int fd = fileno();
    
    if(ioctl(fd, TIOCGLTC, (char*) &Old_ltchars) == ERROR)
	wrong("ioctl TIOCGLTC failed.");
    ltchars New_ltchars = Old_ltchars; 
    New_ltchars.t_lnextc = -1;	// Disable ^V
    New_ltchars.t_suspc = -1;	// Disable ^Z
    New_ltchars.t_dsuspc = -1;	// Disable ^Y

    if(ioctl(fd, TIOCGETP, (char*) &Old_sgttyb) == ERROR)
	wrong("ioctl TIOCGETP failed.");
    ospeed = Old_sgttyb.sg_ospeed;
    if(Old_sgttyb.sg_flags & LCASE)
	wrong("Won't run monocase terminal.");
    sgttyb New_sgttyb = Old_sgttyb;
    New_sgttyb.sg_flags &= ~(short) XTABS;  /* clear XTABS */
    New_sgttyb.sg_flags &= ~(short) CRMOD;  /* clear CRMOD */
    New_sgttyb.sg_flags &= ~(short) ECHO;   /* clear ECHO */
    New_sgttyb.sg_flags |=  (short) CBREAK; /* set   CBREAK */

    if(ioctl(fd, TIOCGETC, (char*) &Old_tchars) == ERROR)
	wrong("ioctl TIOCGETC failed.");
    tchars New_tchars = Old_tchars;
    New_tchars = Old_tchars;
    New_tchars.t_intrc = -1;	/* Disable ^C */
//    New_tchars.t_quitc = -1;	/* Disable ^\ */

    if(ioctl(fd, TIOCSETP, (char*) &New_sgttyb) == ERROR)
	wrong("ioctl TIOCSETP failed.");
    if(ioctl(fd, TIOCSLTC, (char*) &New_ltchars) == ERROR)
    {
	(void) ioctl(fd, TIOCSETP, (char*) &Old_sgttyb);
	wrong("ioctl TIOCSLTC failed.");
    }
    if(ioctl(fd, TIOCGETC, (char*) &New_tchars) == ERROR)
    {
	(void) ioctl(fd, TIOCSETP, (char*) &Old_sgttyb);
	(void) ioctl(fd, TIOCSLTC, (char*) &Old_ltchars);
	wrong("ioctl TIOCGETC failed.");
    }
}

// @@@ TTY::~TTY()

TTY::~TTY()
{
    int fd = fileno();
    
    if(ioctl(fd, TIOCSLTC, (char*) &Old_ltchars) == ERROR)
	bug.error("TTY::~TTY", "ioctl TIOCSLTC failed.");
    if(ioctl(fd, TIOCSETP, (char*) &Old_sgttyb) == ERROR)
	bug.error("TTY::~TTY", "ioctl TIOCSETP failed.");
    if(ioctl(fd, TIOCGETC, (char*) &Old_tchars) == ERROR)
	bug.error("TTY::~TTY", "ioctl TIOCGETC failed.");
}

// @@ * class TERMCAP

// @@@ TERMCAP::TERMCAP()

TERMCAP::TERMCAP()
{
    char entry[1024];

    char* term = getenv("TERM");
    if(term == NULL)
	wrong("Environment variable `TERM' must be set.");
    
    if(tgetent(entry, term) != 1)
    {
	if(strlen(term) < 25)
	    wrong(format("Can't find entry for `%s' in termcap.", term));
	else
	    wrong("Useless value of `TERM' environment.");
    }
    TC_columns = tgetnum("co");
    TC_lines = tgetnum("li");

    char *area = buffer;
    if((TC_left = tgetstr("bc", &area)) == NULL)
    {
	if(!tgetflag("bs"))
	    wrong("Cursor is not able to backspace.");
	TC_left = strcpy(area, "\b");
	area += 2;
    }
    if((TC_goto = tgetstr("cm", &area)) == NULL)
	wrong("Terminal have no cursor addressing.");
    if((TC_clear = tgetstr("cl", &area)) == NULL)
	wrong("Terminal can't clear the screen.");
    TC_line = tgetstr("ce", &area);
    TC_beep = tgetstr("bl", &area);
    TC_region = tgetstr("cs", &area);
    TC_endVisual = tgetstr("ve", &area);
    TC_endCursor = tgetstr("te", &area);
    tputs(tgetstr("ti", &area));
    tputs(tgetstr("vs", &area));
    UP = tgetstr("up", &area);
    char* pad = tgetstr("pc", &area);
    PC = pad ? *pad : 0;
    BC = TC_left;
}
	
// pc - PC = pc pad char

// @@@ TERMCAP::~TERMCAP()

TERMCAP::~TERMCAP()
{
    tputs(TC_endVisual);
    tputs(TC_endCursor);
}

// @@ * class SCREEN

// @@@ char SCREEN::screen(coor pos, const PICTURE& window, coor wpos)

char SCREEN::screen(coor pos, const PICTURE& window, coor wpos)
{
    if(!OnScreen(pos))
	return ' ';
    if(!OnWindow(pos, window, wpos))
	return picture[pos];

    return window[pos - wpos];
}

// @@@ bool SCREEN::IsPictureBlank(const PICTURE& picture,

bool SCREEN::IsPictureBlank(const PICTURE& picture, coor from, coor to)
{
    from = coor(max(0, from.x), max(0, from.y));
    to = coor(min(picture.size().x, to.x + 1),
	      min(picture.size().y, to.y + 1));

    if(from < to)
    {
	for(region pos(from, to); pos.check(); pos.NextX())
	    if(picture[pos] != ' ')
		return FALSE;
    }
    return TRUE;
}    

// @@@ bool SCREEN::IsScreenBlank(coor from, coor to)

bool SCREEN::IsScreenBlank(coor from, coor to)
{
    coor one(max(0, min(from.x, to.x)),
	     max(0, min(from.y, to.y)));
    coor two(min(size().x, max(to.x, from.x) + 1),
	     min(size().y, max(to.y, from.x) + 1));

    if(one < two)
    {
	for(region pos(one, two); pos.check(); pos.NextX())
	    if(screen(pos) != ' ')
		return FALSE;
    }
    return TRUE;
}

// @@@ * void SCREEN::plot(coor pos, char c)

void SCREEN::plot(coor pos, char c)
{
    if(!isprint(c))
	bug.error("SCREEN::plot", format("Wrong char %d", c));
    else if(!OnScreen(pos))
	/* Do nothing. */;
    else if(picture[pos] != c)
    {
	if(!OnWindow(pos) && !OnCorner(pos))
	{
	    if(pos.y > RegionLast || (pos.y == RegionLast && OnEdge(pos)))
		SetScrollRegion(0, size().y - 1);
	    cursor(pos);
	    OutPut(c);
	    cpos.x++;
	    if(OnEdge(pos))
		LoseCursor();
	}
	picture[pos] = c;
    }
}

// @@@ * void SCREEN::print(coor pos, const char* s)

void SCREEN::print(coor pos, const char* s)
{
    while(*s != 0)
    {
	plot(pos, *s);
	s++;
	pos.x++;
    }
}

// @@@ * void SCREEN::ClearLine(coor from, coor to)
//
// Clear the single line from, to.

void SCREEN::ClearLine(coor from, coor to)
{
    // Is it necessary to do anything at all?
    if(IsPictureBlank(picture, from, to) || !OnScreen(coor(0, from.y)))
	return;
    if(from.x < 0)
	from.x = 0;
    to = coor(min(to.x, size().x), from.y);
    // Shall we use the TERMCAP ClearLine?
    if(CanClearLine() &&
       from.x < to.x - 3 &&
       IsPictureBlank(window, from - wpos, to - wpos) &&
       !IsScreenBlank(from, to) &&
       IsScreenBlank(to, coor(size().x, from.y)))
    {
	// Is is necessary to move the cursor?
	if(!OnScreen(cpos) ||
	   from.y != cpos.y ||
	   !IsScreenBlank(from, cpos))
	{
	    cursor(from);
	}
	// Clear the line
	TERMCAP::ClearLine();
	LoseCursor();
	for(coor p = from; p.x < to.x; p.x++)
	    picture[p] = ' ';
    }
    else
    {
	for(coor p = from; p.x < to.x; p.x++)
	    plot(p, ' ');
    }
}

// @@@ * void SCREEN::cursor(coor pos)

void SCREEN::cursor(coor pos)
{
    if(!OnScreen(pos) || pos == cpos)
	return;
    if(!OnScreen(cpos))
	SetCursor(pos);
    else if(pos.y == cpos.y)
    {
	if(pos.x == 0 && cpos.x > 1)
	{
	    OutPut('\r');
	    cpos.x = 0;
	}
	else
	{
	    int dx = pos.x - cpos.x;

	    if(dx > 2 || dx < -2)
		SetCursor(pos);
	    else
	    {
		switch(dx)
		{
		case 2:
		    OutPut(screen(cpos));
		    cpos.x++;
		    /* Fall trough... */
		case 1:
		    OutPut(screen(cpos));
		    break;
		case -2:
		    BackSpace();
		    /* Fall trough... */
		case -1:
		    BackSpace();
		    break;
		default:
		    bug.error("SCREEN::cursor",
			  format("Impossible value for dx = %d", dx));
		}
		cpos = pos;
	    }
	}
    }
    else if(pos == cpos + coor(0, 1) &&
	    RegionFirst <= cpos.y && RegionLast >= pos.y)
    {
	OutPut('\n');
	cpos.y++;
    }
    else
	SetCursor(pos);
}

// @@@ * void SCREEN::ClearScreen()

void SCREEN::ClearScreen()
{

    long cost = 0;		// Calculate cost to redraw window
    coor p;

    for(p.y = 0; p.y < window.size().y; p.y++)
    {
	int first = ERROR;
	for(p.x = 0; p.x < window.size().x; p.x++)
	    if(window[p] != ' ')
	    {
		first = p.x;
		break;
	    }
	if(first != ERROR)
	{
	    for(p.x = window.size().x - 1; window[p] == ' '; p.x--)
		/* Do nothing. */;
	    cost += MagicLineCost + p.x - first + 1;
	}
    }
    if(cost > MagicWindowCost * size().y + size().x)
    {
	for(p = Origo; p.y < size().y; p.y++)
	    ClearLine(p, coor(size().x, p.y));
    }
    else
    {
	picture.resize(size());
	redraw();
    }
}

// @@@ * void SCREEN::ScrollRegion(int first, int last)

void SCREEN::ScrollRegion(int first, int last)
{
    if(first < 0)
	first = 0;
    if(last >= size().y)
	last = size().y;
    if(first >= last)
	return;
    if(CanScrollRegion() &&
       (wpos.y + window.size().y <= first || wpos.y > last || 
	!(window.size() > Origo)))
    {
	if(RegionFirst != first || RegionLast != last)
	    SetScrollRegion(RegionFirst = first, RegionLast = last);
	cursor(coor(0, last));
	OutPut('\n');

	coor p;
	for(p.y = first; p.y < last; p.y++)
	    for(p.x = 0; p.x < size().x; p.x++)
		picture[p] = picture[p + coor(0, 1)];
	for(p = coor(0, last); p.x < size().x; p.x++)
	    picture[p] = ' ';
    }
    else
    {
	coor p;
	for(p.y = first; p.y < last; p.y++)
	    for(p.x = 0; p.x < size().x; plot(p, picture[p + coor(0, 1)]))
		/* Do nothing. */;
	ClearLine(coor(0, last), coor(size().x, last));
    }
}    

// @@@ * void SCREEN::redraw()

void SCREEN::redraw()
{
    TERMCAP::ClearScreen();
    LoseCursor();

    coor p;
    for(p.y = 0; p.y < size().y; p.y++)
	for(p.x = 0; p.x < size().x; p.x++)
	    if(screen(p) != ' ' && !OnCorner(p))
	    {
		cursor(p);
		OutPut(screen(p));
		if(OnEdge(p))
		    LoseCursor();
		else
		    cpos.x++;
	    }
}

// @@@ * void SCREEN::cover(const PICTURE& window, coor pos)

void SCREEN::cover(const PICTURE& cwin, coor pos)
{
    coor from(min(pos.x, wpos.x), min(pos.y, wpos.y));
    coor to(max(size().x,
		max(pos.x + cwin.size().x,
		    wpos.x + window.size().x)),
	    max(size().y,
		max(pos.y + cwin.size().y,
		    wpos.y + window.size().y)));
    coor p;
    
    for(p.y = from.y; p.y < to.y; p.y++)
    {
	int tail = size().x;

	if(CanClearLine())
	{
	    while(tail > from.x)
	    {
		if(screen(coor(tail - 1, p.y), cwin, pos) != ' ')
		    break;
		tail--;
	    }
	    if(tail > size().x - 3)
		tail = size().x;
	}
	for(p.x = from.x; p.x < min(to.x, tail); p.x++)
	{
	    const int c = screen(p, cwin, pos);
	    
	    if(c != screen(p) && !OnCorner(p))
	    {
		cursor(p);
		OutPut(c);
		if(OnEdge(p))
		    LoseCursor();
		else
		    cpos.x++;
	    }
	}
	if(tail < size().x &&
	   !IsScreenBlank(coor(tail, p.y), coor(size().x - 1, p.y)))
	{
	    cursor(coor(tail, p.y));
	    TERMCAP::ClearLine();
	    LoseCursor();
	}
    }
    window = cwin;
    wpos = pos;
}    

// @@@ * void SCREEN::cover()

void SCREEN::cover()
{
    const PICTURE window;
    cover(window, Origo);
}

// @@@ SCREEN::SCREEN()

SCREEN::SCREEN()
{
    picture.resize(size());
    LoseCursor();
    wpos = Origo;
    RegionFirst = 0;
    RegionLast = size().y - 1;
    redraw();
}

// @@ data

static pfVpcC screen_handler = NULL;
static pfVC output_char = NULL;

// @@ function

// @@@ static void wrong(const char* what)

static void wrong(const char* what)
{
    bug.error("Screen Handler", what);
    if(screen_handler)
	screen_handler(what);
    else
	(void) fprintf(stderr, "Sorry: %s\r\n", what);
    exit(1);
}

// @@@ static void wrong(STRING what)

static void wrong(STRING what)
{ wrong(what()); }

// @@@ * pfVpcC set_screen_handler(pfVpcC screen_handler)

pfVpcC set_screen_handler(pfVpcC screen_handler)
{
    pfVpcC old = ::screen_handler;
    ::screen_handler = screen_handler;
    return old;
}

// @@@ static void OutPut(char c)

static void OutPut(char c)
{
    if(output_char)
	output_char(c);
    else
	(void) putchar(c);
}

// @@@ void set_output_char(pfVC output_char)

void set_output_char(pfVC output_char)
{ ::output_char = output_char; }

// @@@ * void tputs(const char* cp, int affcnt = 1)

void tputs(const char* cp, int affcnt = 1)
{ (void) tputs((char*) cp, affcnt, (TPUTS_FUN) OutPut); }

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
