// @ myth - server - pack.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/pack.h"
#include "server/bunch.h"
#include "server/player.h"
#include "common/error.h"
#include "server/rnd.h"
#include "server/specify.h"

// @@ PACK

// @@@ void PACK::RemoveHalf(SPECIFY* specify, 
//
// Just like remove, except that is don't touch worn_list

void PACK::RemoveHalf(SPECIFY* specify, OBJECT_LIST* list, long argument = 0L)
{
    if(argument > 0L)
    {
	long total = 0L;

	reset();
	while(current() != NULL)
	{
	    if(specify->check(current()))
		total += current()->count();

	    next();
	}
	if(total > argument)
	{
	    reset();
	    while(current() != NULL && argument > 0)
	    {
		long count = current()->count();
		total -= count;

		if(specify->check(current()) &&
		   rnd.chance((int) argument, (int) total + 1))
		{
		    if(argument >= count)
		    {
			list->append(current());
			argument -= count;
			INVENTORY::RemoveObject(); // Leave worn_list alone
		    }
		    else
		    {
			list->append(current()->split(argument));
			return;
		    }
		}
		else
		    next();
	    }
	    return;
	}
    }
    reset();
    while(current() != NULL)
    {
	if(specify->check(current()))
	{
	    list->append(current());
	    INVENTORY::RemoveObject(); // I said "Dont touch worn_list!"
	}
	else
	    next();
    }
}

// @@@ void PACK::DropDead(BUNCH* bunch)

void PACK::DropDead(BUNCH* bunch)
{
    reset();
    while(current() != NULL)
    {
	// Most worn objects gets cursed, most other don't
	if((worn(current()) && rnd.chance(9,10)) || rnd.chance(1,10))
	    current()->curse();
	    
	// Drop it on the floor
	bunch->add(current());
	RemoveObject();
    }
}

// @@@ OBJECT* PACK::FindWorst(MONSTER* monster)
//
// Return the object from the pack which has the lowest value to the monster

OBJECT* PACK::FindWorst(MONSTER* monster)
{
    OBJECT* worst = NULL;

    reset();
    while(current() != NULL)
    {
	if(current()->CanRemove(monster) &&
	   (worst == NULL ||
	    current()->worth(monster) < worst->worth(monster)))
	{
	    worst = current();
	}
	next();
    }
    return worst;
}

// @@@ void PACK::RemoveObject()

void PACK::RemoveObject()
{
    worn_list.remove(current());
    INVENTORY::RemoveObject();
}

// @@@ void PACK::AddObject(OBJECT* object)

void PACK::AddObject(OBJECT* object)
{
    INVENTORY::AddObject(object);
}

// @@@ bool PACK::worn(OBJECT* object)

bool PACK::worn(OBJECT* object)
{ return worn_list.member(object); }

// @@@ void PACK::wear(OBJECT* object)

void PACK::wear(OBJECT* object)
{ worn_list.add(object); }

// @@@ void PACK::TakeOff(OBJECT* object)

void PACK::TakeOff(OBJECT* object)
{ worn_list.remove(object); }

// @@@ OBJECT* PACK::weapon()

OBJECT* PACK::weapon()
{
    reset();
    while(current() != NULL)
    {
	if(current()->symbol() == SYMBOL_WEAPON && worn(current()))
	    return current();

	next();
    }
    return NULL;
}

// @@@ OBJECT* PACK::armor()

OBJECT* PACK::armor()
{
    reset();
    while(current() != NULL)
    {
	if(current()->symbol() == SYMBOL_ARMOR && worn(current()))
	    return current();

	next();
    }
    return NULL;
}

// @@@ void PACK::destroy(OBJECT* object)

void PACK::destroy(OBJECT* object)
{
    touch();
    junk.add(object);
}

// @@@ void PACK::SendPack(PLAYER* player, force flag = CHECK)

void PACK::SendPack(PLAYER* player, force flag = CHECK)
{
    if(TouchPack || flag == FORCE)
    {
	TouchPack = FALSE;
	send(INVENTORY_OWN, player);
    }
}

// @@@ void PACK::update(LEVEL* level, coor pos, MONSTER* monster)

void PACK::update(LEVEL* level, coor pos, MONSTER* monster)
{
    // Clear old values of abilities.
    ability.clear();
    NumberOfRings = 0;

    // Create new abilities.
    reset();
    while(current() != NULL)
    {
	if(current()->IsRing() && worn(current()))
	    NumberOfRings++;
	
	current()->InsertAbility(monster);
	next();
    }

    // Update the pack.
    INVENTORY::update(level, pos, monster);
    TouchPack = TRUE;
}

// @@@ damage_type PACK::absorb(int* power,

damage_type PACK::absorb(int* power, OBJECT* object,
			 MONSTER* attack, MONSTER* defence)
{
    damage_type type = object->DamageType();

    reset();
    while(remain() > 0)
    {
        if(current()->absorb(power, &type, object, attack, defence) == DONE)
	{
            destroy(current());
	    RemoveObject();
	}
	else
	    next();
    }
    return type;
}

// @@@ void PACK::create()

void PACK::create()
{ TouchPack = TRUE; }

// @@@ void PACK::save(BINARY* binary)

void PACK::save(BINARY* binary)
{
    // Get rid of the junk
    junk.reset();
    while(junk.length() > 0)
	junk.erase();

    // Save the important stuff
    worn_list.save(binary);
    INVENTORY::save(binary);
}

// @@@ void PACK::load(BINARY* binary)

void PACK::load(BINARY* binary)
{
    worn_list.load(binary);
    INVENTORY::load(binary);
}

// @@@ PACK::PACK()

PACK::PACK() : ()
{ create(); }

// @@@ PACK::PACK(PARSE* parse, coor pos)

PACK::PACK(PARSE* parse, coor pos) : (parse, pos)
{ create(); }

// @@@ PACK::~PACK()

PACK::~PACK()
{ }

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
